#define DEBUGGING 1
/*
 * linux/arch/arm/kernel/head32.S
 *
 * Copyright (C) 1994, 1995, 1996, 1997 Russell King
 *
 * Kernel 32 bit startup code for ARM6 / ARM7 / StrongARM
 */
#include <linux/config.h>
#include <asm/hardware.h>

		.text
		.align

		.globl	swapper_pg_dir
		.equ	swapper_pg_dir,	0xc0008000

		.globl 	_start
		.globl	_entry
		.globl	_stext
/*
 * Entry point and restart point.  Entry *must* be called with r0 == 0,
 * MMU off.
 *
 *  r1 = 0 -> ebsa  (Ram @ 0x00000000)
 *  r1 = 1 -> RPC   (Ram @ 0x10000000)
 *  r1 = 2 -> ebsit (???)
 */

#define DC21285_ARMCSR_BASE             0x42000000
#define CSR_BASE                        DC21285_ARMCSR_BASE

#define IRQ_ENABLE_CLEAR                0x18C
#define FIQ_ENABLE_CLEAR                0x28C

#define XBUS_DEV_2                      DC21285_XBUS_XCS2

#define LED_BANK                        XBUS_DEV_2
#define RED_LED                         0x04
#define GREEN_LED                       0x02
#define YELLOW_LED                      0x01

#define IO_BASE_PHYSICAL                0x7c000000
#define IO_BASE_VIRTUAL                 0xE0000000

#define PAGE_TABLE_ROOT                 0x00004000
_stext:
_start:
_entry:
@
@ Tell the world we've arrived
@

@advance PC to start+0x20, as the loader jumps to base+0x20

		nop
		nop
		nop
		nop

		nop
		nop
		nop
		nop
		

                BL      resetCOM

		adr	r0, Lstring_e
		bl	pputs
		b	1f

Lstring_e:	.ascii  "\n\re.\0"
Lstring_n:	.ascii  "n.\0"
Lstring_rev:    .ascii  "rev:\0"
Lstring_t:	.ascii  "t.\0"
Lstring_y:	.ascii  "y. \0"
		.align
		.global board_id
board_id:	.word 0
		.align

@ ----------------------------------------------------------------------
@ Deal with being loaded by Angel/Demon
@
@
@
@ Disable clock switching
@
1:              mcr     p15, 0, r0, c15, c2 ,2

@
@ Turn off interrupts from the 21285
@
                ldr     r0, =0xFFFFFFFF
                ldr     r4, =CSR_BASE
                str     r0, [r4, #IRQ_ENABLE_CLEAR]
                str     r0, [r4, #FIQ_ENABLE_CLEAR]

@
@ disable the i-cache
@
                mov     r0, #0                  @ Disable MMU & caches
                mcr     p15, 0, r0, c1, c0
                mcr     p15, 0, r0, c8, c7      @ flush I,D TLBs
                mcr     p15, 0, r0, c7, c7      @ flush I,D caches
                nop                             @ make sure that there are
                nop                             @ no pipelined instructions
                nop
                nop

		adr     r0, Lstring_n
		bl	pputs
@
@ NetWinder: determine board rev
@        pre-MMU, so the addresses are 21285 physical 

		@ point into empty PCI space
		mov 	r4, #DC21285_PCI_MEMORY @ empty pci memory address
		@ wait for the bus to settle
		bl 	pdelay
		@ read the board rev, bits 15..8
		ldrh 	r5, [r4]
		@ store the board rev in 
		adr 	r4, board_id
		str 	r5, [r4]

@
@ mapping physical memory
@ Map physical memory to virtual for the first 8Mbytes
@
#ifndef CONFIG_ARCH_VNC
                mov     r0, r4
                orr     r1, r5, #0x00000002     @ Physical memory | TYPE_SECT
                orr     r1, r1, #0x00000c00     @ AP_READ | AP_WRITE
                str     r1, [r0], #4
                add     r1, r1, #1 << 20
                str     r1, [r0], #4
                add     r1, r1, #1 << 20
                str     r1, [r0], #4
                add     r1, r1, #1 << 20
                str     r1, [r0], #4
                add     r1, r1, #1 << 20
                str     r1, [r0], #4
                add     r1, r1, #1 << 20
                str     r1, [r0], #4
                add     r1, r1, #1 << 20
                str     r1, [r0], #4
                add     r1, r1, #1 << 20
                str     r1, [r0], #4
#endif

		adr	r0, Lstring_t
		bl	pputs

@
@ end of angel/demon stuff
@ ----------------------------------------------------------------------
@
@ Fudge the types (for now)
@
                mov     r0, #0
                mov     r1, #0
@
@ First thing to do is to get the page tables set up so that we can call the kernel
@ in the correct place.  This is relocatable code...
@
		mrc	p15, 0, r9, c0, c0			@ get Processor ID
@
@ Read processor ID register (CP#15, CR0).
@ NOTE: ARM2 & ARM250 cause an undefined instruction exception...
@ Values are:
@		XX01XXXX = ARMv4 architecture (StrongARM)
@		XX00XXXX = ARMv3 architecture
@		4156061X = ARM 610
@ 		4156030X = ARM 3
@		4156025X = ARM 250
@		4156020X = ARM 2
@
		adr	r10, LCProcTypes
1:		ldmia	r10!, {r5, r6, r8}			@ Get Set, Mask, MMU Flags
		teq	r5, #0					@ End of list?
		beq	Lerror
		eor	r5, r5, r9
		tst	r5, r6
		addne	r10, r10, #8
		bne	1b

		adr	r4, LCMachTypes
		add	r4, r4, r1, lsl #4
		ldmia	r4, {r4, r5, r6}			@ r4 = page dir in physical ram
                                                                @ r5 = physical memory
                                                                @ r6 = io physical location

		mov	r0, r4
		mov	r1, #0
		add	r2, r0, #0x4000
1:		str	r1, [r0], #4				@ Clear page table
		teq	r0, r2
		bne	1b

@
@ Add enough entries to allow the kernel to be called.
@ It will sort out the real mapping in paging_init
@
@ Map the physical memory to va 0xC0000000
@
                @ load the first address to burn
                adr     r7, LCMachTypes
                ldr     r4, [r7], #4    @ address of page table
                ldr     r5, [r7], #4    @ address of base ram
                add     r0, r4, #0x3000 @ map it to 0xC0000000
		mov	r1, #0x0000000c				@ SECT_CACHEABLE | SECT_BUFFERABLE
		orr	r1, r1, r8
		add	r1, r1, r5
		str	r1, [r0], #4
		add	r1, r1, #1 << 20
		str	r1, [r0], #4
		add	r1, r1, #1 << 20
@
@ Map in IO space (to va 0xD0000000)
@
		add	r0, r4, #0x3400
mmu_init_loop:
                ldr     r6, [r7], #4
                cmp     r6, #0
                beq     mmu_init_done

		orr	r1, r6, r8

                ldr     r2, [r7], #4            @ length of segment
                add     r2, r2, r1              @ end of segment
1:		str	r1, [r0], #4
		add	r1, r1, #1 << 20
		teq	r1, r2
		bne	1b

                mov     r5, r0
		adr	r0, Lstring_r
		bl	pputs
                mov     r0, r5

                b       mmu_init_loop

mmu_init_done:
		adr	r0, Lstring_y
		bl	pputs
 
@
@ Go to red, don't touch r4, that contains the page table address in physical memory
@
                LDR     r5, =RED_LED
                EOR     r5, r5, #7
                LDR     r6, =LED_BANK
                STR     r5, [r6]

@
@ The following should work on both v3 and v4 implementations
@
		mov	lr, pc
		mov	pc, r10					@ Call processor flush (returns ctrl reg)
		adr	r5, _entry
		sub	r10, r10, r5				@ Make r10 PIC

		ldr	lr, Lbranch
		mcr	p15, 0, r0, c1, c0			@ Enable MMU & caches.  In 3 instructions
								@ we loose this page!
		mov	pc, lr

Lerror:		mov	r0, #0x02000000
		mov	r1, #0x11
		orr	r1, r1, r1, lsl #8
		orr	r1, r1, r1, lsl #16
		str	r1, [r0], #4
		str	r1, [r0], #4
		str	r1, [r0], #4
		str	r1, [r0], #4
		b	Lerror

Lbranch:	.long	Lalready_done_mmap			@ Real address of routine

                @ VNC  --- actually, all we really need is the first one, the IO space, because the C memory mapper rewrites it all anyway
LCMachTypes:    .long   swapper_pg_dir - 0xC0000000
		.long	0					@ Address of RAM
                .long   0x88000000
                .long   SZ_256M
                .long   DC21285_PCI_IO  @ PCI io space 7c00000
                .long   SZ_1M
                .long   DC21285_PCI_TYPE_0_CONFIG
                .long   0x00F00000      @ SZ_15M
                .long   DC21285_ARMCSR_BASE
                .long   SZ_1M
                .long   DC21285_PCI_IACK
                .long   SZ_1M
                .long   DC21285_SA_CACHE_FLUSH
                .long   SZ_1M
                .long   DC21285_OUTBOUND_WRITE_FLUSH
                .long   SZ_1M
                .long   DC21285_ETH100_MEM
                .long	SZ_1M
                .long   DC21285_DRAM_A0MR
                .long   SZ_1M + SZ_2M
                .long   EPROM_BASE
                .long   SZ_8M
                .long   0
 
                @ EBSA285 (pg dir phys, phys ram start, phys i/o)
                .long   swapper_pg_dir - 0xc0000000
                .long   0
                .long   IO_BASE_PHYSICAL
		.long	0

		@ RPC
		.long	swapper_pg_dir - 0xc0000000 + 0x10000000
		.long	0x10000000
		.long	0x03000000
		.long	0

		@ EBSIT ???
		.long	swapper_pg_dir - 0xc0000000
		.long	0
		.long	0xe0000000
		.long	0

LCProcTypes:	@ ARM6 / 610
		.long	0x41560600
		.long	0xffffff00
		.long	0x00000c12
		b	Larmv3_flush_early			@ arm v3 flush & ctrl early setup
		mov	pc, lr

		@ ARM7 / 710
		.long	0x41007000
		.long	0xfffff000
		.long	0x00000c12
		b	Larmv3_flush_late			@ arm v3 flush & ctrl late setup
		mov	pc, lr

		@ StrongARM
		.long	0x4401a100
		.long	0xfffffff0
		.long	0x00000c02
		b	Larmv4_flush_early
		b	Lsa_fastclock

Lstring_r:	.ascii  "r.\0"
Lstring_d:      .ascii  "d.\0"
Lstring_o:      .ascii  "one.\0"
		.align

		.long	0

LC0:		.long	_edata
		.long	arm_id
		.long	_end
		.long	init_user_stack + 4096
		.align

Larmv3_flush_early:
		mov	r0, #0
		mcr	p15, 0, r0, c7, c0			@ flush caches on v3
		mcr	p15, 0, r0, c5, c0			@ flush TLBs on v3
		mcr	p15, 0, r4, c2, c0			@ load page table pointer
 		mov	r0, #5					@ Domains 0, 1 = client
		mcr	p15, 0, r0, c3, c0			@ load domain access register
		mov	r0, #0x3d				@ ....DPWCAM
		mov	pc, lr

Larmv3_flush_late:
		mov	r0, #0
		mcr	p15, 0, r0, c7, c0			@ flush caches on v3
		mcr	p15, 0, r0, c5, c0			@ flush TLBs on v3
		mcr	p15, 0, r4, c2, c0			@ load page table pointer
		mov	r0, #5					@ Domains 0, 1 = client
		mcr	p15, 0, r0, c3, c0			@ load domain access register
		mov	r0, #0x7d				@ ...LDPWCAM
		mov	pc, lr

Larmv4_flush_early:
		mov	r0, #0
		mcr	p15, 0, r0, c7, c7			@ flush I,D caches on v4
		mcr	p15, 0, r0, c7, c10, 4			@ drain write buffer on v4
		mcr	p15, 0, r0, c8, c7			@ flush I,D TLBs on v4
		mcr	p15, 0, r4, c2, c0			@ load page table pointer

		mov	r0, #5					@ Domains 0, 1 = client
		mcr	p15, 0, r0, c3, c0			@ load domain access register

                @ check version of the footbridge
                mov     r1, #0x42000000
                ldrb    r1, [r1, #8]            @ get rev level
                cmp     r1, #0
@ no alignment trap setting		
@                movne   r1, #0x3d               @ ....DPWC.M
@                moveq   r1, #0x35               @ ....DP.C.M (no write buffer if the footbridge is rev 0)
@alignment trap stuff
                movne   r1, #0x3f               @ ....DPWCAM
                moveq   r1, #0x37               @ ....DP.CAM (no write buffer if the footbridge is rev 0)

                @ enable caches
                mrc     p15, 0, r0, c1, c0      @ get control register v4
                orr     r0, r0, r1
                orr     r0, r0, #1 << 12        @ v4 supports separate I cache
		mov	pc, lr

Lsa_fastclock:	mcr	p15, 0, r4, c15, c1, 2			@ Enable clock switching
		mov	pc, lr

Lalready_done_mmap:
@
@ Go to YELLOW (use virtual addresses)
@
		adr	r0, Lstring_d
		bl	vputs

		adr	r5, _entry				@ Add base back in
		add	r10, r10, r5

		adr	r5, LC0
		ldmia	r5, {r5, r6, r8, sp}			@ Setup stack
		mov	r4, #0
1:		cmp	r5, r8					@ Clear BSS
		strcc	r4, [r5],#4
		bcc	1b


		adr	r0, Lstring_o
		bl	vputs

		mov	lr, pc
		add	pc, r10, #4				@ Call post-processor init

		mov	fp, #0

		b	start_kernel

@ debug stuff HACK
@ UART registers

#define COM1Port                0x3f8
#define COM2Port                0x2f8

#define DC21285_XBUS_XCS1       0x40011000

#define XBUS_DEV_1              DC21285_XBUS_XCS1

#ifdef CONFIG_ARCH_VNC

#undefine COM1
#undefine COM2
#define COM1                   COM1Port + IO_BASE_PHYSICAL
#define COM2                   COM2Port + IO_BASE_PHYSICAL


#else
#define IOShift                 2
#define        COM1                    (XBUS_DEV_1 + (COM1Port << IOShift))
#define COM2                   (XBUS_DEV_1 + (COM2Port << IOShift))
#endif
#define Rx                     0
#define Tx                     0

#ifdef CONFIG_ARCH_VNC
#define IntEnable              1
#define FIFOcntl               2
#define LineCntl               3
#define ModemCntl              4
#define LineStatus              5
#define Dllsb                  0
#define Dlmsb                  4
#else
#define IntEnable              0x4
#define FIFOcntl               0x8
#define LineCntl               0xC
#define ModemCntl              0x10
#define LineStatus              0x14
#define Dllsb                  0
#define Dlmsb                  1
#endif

#define DLABMsk                        0x80
#define LineTEMTMsk            0x40

#define Baud9600low            12
#define Baud9600high           0
#define Baud19200low           6
#define Baud19200high          0

#define WordLen8               3



               .globl  _resetCOM
resetCOM:      LDR     r2,=COM2
               ldr     r3,=COM1
#ifdef CONFIG_ARCH_VNC
                mov     r1, #0                  @disable ints
                orr     r1, r1, #DLABMsk << 24  @ set for divisor access
                str     r1, [r2]
                str     r1, [r3]
                mov     r1, #Baud19200low       @set baud rate
                orr     r1, r1, #Baud19200high << 8
                orr     r1, r1, #WordLen8 << 24 @and lock in the baud rate
                str     r1, [r2]
                str     r1, [r3]
                ldr     r1, [r2, #4]            @ clear line status
                ldr     r1, [r3, #4]
#else
                ldr     r1, =0                  @ disable interrupt
                strb    r1, [r2, #IntEnable]
                ldr     r1, =DLABMsk            @ Set divisor access
                strb    r1, [r2, #LineCntl]
                ldr     r1, =Baud19200low       @ Set 19200 low divisor
                strb    r1, [r2, #Dllsb]
                ldr     r1, =Baud19200high      @ set 19200 high divisor
                strb    r1, [r2, #Dlmsb]
                ldr     r1, =WordLen8           @ Clear divisor access
                strb    r1, [r2, #LineCntl]
                ldr     r1, =0                  @ 0 disables the FIFOs
                strb    r1, [r2, #FIFOcntl]
                ldrb    r1, [r2, #LineStatus]   @ read status will clear errors
#endif
                ldr     r1, =0xf
                strb    r1, [r2, #ModemCntl]    @ disable modem control
                strb    r1, [r3, #ModemCntl]    @ disable modem control
                 MOV     pc, lr
 
                .globl  _printchV       @ virtual
 _printchV:     ldr     r3, =(COM2 - IO_BASE_PHYSICAL + IO_BASE_VIRTUAL)
                b       _printcharacter
                .globl  _printchP       @ physical
_printchP:      ldr     r3, =COM2
_printcharacter:
#ifdef CONFIG_ARCH_VNC
4:              ldr     r2,[r3,#4]
                tst     r2,#LineTEMTMsk << 8         @ Wait until ready
#else
4:              ldrb    r2,[r3,#LineStatus]
                tst     r2,#LineTEMTMsk         @ Wait until ready
#endif
                beq     4b
                strb    r0,[r3,#Tx]
                strb    r0,[r3,#Tx+0x100]       @ !! double console !!
                mov     pc, lr

@
@ Virtual led writing routine
@
                .globl _putLEDs
_putLEDs:
                stmfd   sp!, {r1}
                eor     r0, r0, #7
                ldr     r1, =(LED_BANK - IO_BASE_PHYSICAL + IO_BASE_VIRTUAL)
                str     r0, [r1]
                ldmfd   sp!, {r1}
                mov     pc, lr

@
@ Virtual character writing routine
@
                .globl putchar
putchar:
                stmfd   sp!, {r2, r3, lr}
                bl      _printchV
                ldmfd   sp!, {r2, r3, pc}
#if 0 
                .globl puthexP
puthexP:

                stmfd   sp!, {r2, r3, r4, lr}
                mov     r4, r0

                mov     r0, r4, lsr #28
                cmp     r0, #9
                addle   r0, r0, #48
                addgt   r0, r0, #55
                bl      _printchP

                mov     r0, r4, lsr #24
                and     r0, r0, #0xF
                cmp     r0, #9
                addle   r0, r0, #48
                addgt   r0, r0, #55
                bl      _printchP

                mov     r0, r4, lsr #20
                and     r0, r0, #0xF
                cmp     r0, #9
                addle   r0, r0, #48
                addgt   r0, r0, #55
                bl      _printchP

                mov     r0, r4, lsr #16
                and     r0, r0, #0xF
                cmp     r0, #9
                addle   r0, r0, #48
                addgt   r0, r0, #55
                bl      _printchP

                mov     r0, r4, lsr #12
                and     r0, r0, #0xF
                cmp     r0, #9
                addle   r0, r0, #48
                addgt   r0, r0, #55
                bl      _printchP

                mov     r0, r4, lsr #8
                and     r0, r0, #0xF
                cmp     r0, #9
                addle   r0, r0, #48
                addgt   r0, r0, #55
                bl      _printchP

                mov     r0, r4, lsr #4
                and     r0, r0, #0xF
                cmp     r0, #9
                addle   r0, r0, #48
                addgt   r0, r0, #55
                bl      _printchP

                mov     r0, r4
                and     r0, r0, #0xF
                cmp     r0, #9
                addle   r0, r0, #48
                addgt   r0, r0, #55
                bl      _printchP

                ldmfd   sp!, {r2, r3, r4, pc}
#endif

	@ just a quick delay about 1us
	.global pdelay
pdelay:	mov r0,#512
1:	subs r0,r0,#1
	bne 1b
	mov pc,lr

	@ a string putter
	.global pputs
pputs:
	mov r1, r0
1:	ldrb r0, [r1], #1
	cmp r0, #0
	moveq pc, lr
	mov ip, lr
	bl _printchP
	mov lr, ip
	b 1b 

	@ and the same for virtual addresses
	.global vputs
vputs:	mov r1, r0
1:	ldrb r0, [r1], #1
	cmp r0, #0
	moveq pc, lr
	mov ip, lr
	bl _printchV
	mov lr, ip
	b 1b

#ifdef DEBUGGING
@ print a hex char in r0
@ uses ip, r8, r9, and then r1-3 through the call to _printchP
@ for use before the mmu is turned on
pputhex:
	mov ip, lr
        mov r8, #8
        mov r9, r0
        mov r0, #'0'
        bl  _printchP
        mov r0, #'x'
        bl  _printchP
2:      mov r0, r9, lsr #28
        add r0, r0, #'0'
        cmp r0, #'9'
        ble 1f
        add r0, r0, #7
1:	bl  _printchP
	movs r9, r9, lsl #4
        subs r8, r8, #1
	bne 2b
        mov pc, ip
#endif
