/*
 * linux/include/asm-arm/io.h
 *
 * Copyright (C) 1996 Russell King
 *
 * Modifications:
 *  16-Sep-1996	RMK	Inlined the inx/outx functions & optimised for both
 *			constant addresses and variable addresses.
 * !! Much of this stuff is VNC specific, so should move to include/asm-arm/arch-vnc/io.h
 */
#ifndef __ASM_ARM_IO_H
#define __ASM_ARM_IO_H

#include <asm/hardware.h>
#include <asm/arch/mmu.h>

#define virt_to_bus(x)  ((unsigned long)((unsigned long)x - 0xC0000000))
#define bus_to_virt(x)  ((void*)((unsigned long)x + 0xC0000000))

/*
 * IO -> physical
 */

extern __inline__ void __outb(unsigned int value, unsigned int port)
{
	volatile int i;
	i=0; i=0; // delay
        *(volatile char*)(IO_BASE + port) = value;
}

//extern __inline__ void __outw(unsigned int value, unsigned int port)
//{
//        __outb(value & 0xFF, port);
//        __outb(value >> 8, port);
//}
extern __inline__ void __outw(unsigned int value, unsigned int port)
{
	port += IO_BASE;

__asm__("
	mov	r0,%0
	mov	r1,%1
	.word	0xE04100B0
	" : : "r"(value),"r"(port) : "r0","r1");
}
	
extern __inline__ void __outl(unsigned int value, unsigned int port)
{
        *(volatile int*)(IO_BASE + port) = value;
}

extern __inline__ unsigned char __inb(unsigned int port)
{
        return *(volatile char*)(IO_BASE + port);
}

//extern __inline__ unsigned short __inw(unsigned int port)
//{
//        unsigned long value;
//        value = __inb(port);
//        value |= __inb(port) << 8;
//	return value;
//}
extern __inline__ unsigned short __inw(unsigned int port)
{
unsigned short s;

	port += IO_BASE;

__asm__("
	mov	r1,%1
	.word	0xE05100B0
	mov	%0,r0
	" : "=r"(s) : "r"(port) : "r0","r1");

	return s;

}

extern __inline__ unsigned long __inl(unsigned int port)
{
        return *(volatile unsigned long*)(port + IO_BASE);
}

extern __inline__ unsigned long __ioaddr(unsigned int port)
{
        return (unsigned long)(IO_BASE + port);
}


extern __inline__ void insl(unsigned int port, void *to, int len)
{
        unsigned long *p = (unsigned long*) to;
if ((int)to & 3)
printk("ASSERT: insl misaligned buffer!!!\n");
        while (len--)
                *p++ = *(volatile unsigned long*)(port + IO_BASE);
}

extern __inline__ void insw(unsigned int port, void *to, int len)
{
        unsigned short *p = (unsigned short*) to;
        while (len--)
                *p++ = __inw(port);
}

extern __inline__ void insb(unsigned int port, void *to, int len)
{
        unsigned char *p = (unsigned char*) to;
        while (len--)
                *p++ = *(volatile unsigned char*)(port + IO_BASE);
}
extern __inline__ void outsl(unsigned int port, const void *from, int len)
{
        unsigned long *p = (unsigned long *)from;
        while(len-- > 0)
                *(volatile unsigned long *)(port + IO_BASE) = *p++; 
}
extern __inline__ void outsw(unsigned int port, const void *from, int len)
{
        unsigned short *p = (unsigned short *)from;
        while(len-- > 0)
                __outw(*p++, port);
}
extern __inline__ void outsb(unsigned int port, const void *from, int len)
{
        unsigned char *p = (unsigned char*) from;
        while (len--)
                *(volatile unsigned char*)(port + IO_BASE) = *p++;
}

#define arm_outb(v,p) \
	(*(volatile unsigned char *)(p) = (v))

#define arm_inb(p) \
	(*(volatile unsigned char *)(p))

#define arm_outl(v,p) \
	(*(volatile unsigned long *)(p) = (v))

#define arm_inl(p) \
	(*(volatile unsigned long *)(p))


#define outb_p(val,port) outb((val),(port))
#define outw_p(val,port) outw((val),(port))
#define outl_p(val,port) outl((val),(port))
#define inb_p(port) inb((port))
#define inw_p(port) inw((port))
#define inl_p(port) inl((port))

#define outsb_p(port,from,len) outsb(port,from,len)
#define outsw_p(port,from,len) outsw(port,from,len)
#define outsl_p(port,from,len) outsl(port,from,len)
#define insb_p(port,to,len) insb(port,to,len)
#define insw_p(port,to,len) insw(port,to,len)
#define insl_p(port,to,len) insl(port,to,len)

#define outb(val,port)	__outb((val),(port))
#define outw(val,port)	__outw((val),(port))
#define outl(val,port)	__outl((val),(port))
#define inb(port)	__inb((port))
#define inw(port)	__inw((port))
#define inl(port)	__inl((port))
#define ioaddr(port)	__ioaddr((port))
  

#endif
